(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Regression
imports "../TermFun"
begin

text \<open>simple recursive identity\<close>

function f :: "nat \<Rightarrow> nat"
where
  "f 0 = 0"
| "f (Suc n) = Suc (f n)"
by pat_completeness auto

define_trs (no_sim) f.psimps

termination f
apply (trs_reduce f)
apply expand_trs_problem
export_termination_problem "f.xml"
apply (ceta_import_proof "proofs/f.proof.xml")
apply eval 
done

text \<open>and again, using dp-extraction\<close>

function ff :: "nat \<Rightarrow> nat"
where
  "ff 0 = 0"
| "ff (Suc n) = Suc (ff n)"
by pat_completeness auto

define_dps ff.psimps

termination ff
apply (dps_reduce ff)
apply expand_dp_problem
export_dp_problem "ff.xml"
apply (ceta_import_dp_proof "proofs/ff.proof.xml")
apply eval 
done

define_trs ff.simps (* generate simulation rules for ff *)

text \<open>nested case\<close>

definition predpred :: "nat \<Rightarrow> nat"
where
  "predpred n = (case n of 
    0 \<Rightarrow> 0
  | Suc m \<Rightarrow> case m of 
       0 \<Rightarrow> 0
     | Suc m' \<Rightarrow> m')"

define_trs predpred_def

text \<open>if\<close>

function test1 :: "nat \<Rightarrow> nat"
where
  "test1 n = (if n = 0 then 0 else Suc (test1 (n - Suc 0)))"
by pat_completeness auto

define_trs (no_sim) test1.psimps

termination test1
apply (trs_reduce test1) 
apply expand_trs_problem
export_termination_problem "test1.xml"
apply (ceta_import_proof "proofs/test1.proof.xml") 
apply eval
done

end
